/*
 * out_yasapi-nt.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <windows.h>
#include <shlobj.h>
#include <yasapi.h>


int getwrittentime();
int getoutputtime();

static Player player;
static Out_Module out;

BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
  return TRUE;
}

void config(HWND hwnd)
{
  PlayerShow(&player,&player.gtk.config,hwnd);
}

void about(HWND hwnd)
{
  PlayerShow(&player,&player.gtk.about,hwnd);
}

void init()
{
  if (PlayerCreate(&player,&out)<0) {
    DMESSAGE("creating player");
    goto player;
  }

  return;
//cleanup:
  PlayerDestroy(&player);
player:
  return;
}

void quit()
{
  PlayerDestroy(&player);
}

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
  int latency;

  latency=PlayerOpen(&player,samplerate,numchannels,bitspersamp);

  return latency;
}

void close()
{
  EngineClose(&player.engine);
}

int write(char *buf, int len)
{
  // 0 on success. Len == bytes to write (<= 8192 always). buf is straight
  //   audio data. 
  // 1 returns not able to write (yet). Non-blocking, always.
  if (PlayerWrite(&player,(BYTE *)buf,len)<0)
    return 1;

  return 0;
}

int canwrite()
{
  return PlayerAvailable(&player);
}

int isplaying()
{
  return PlayerIsPlaying(&player);
}

int pause(int pause)
{
  return PlayerPause(&player,pause);
}

void setvolume(int volume)
{
  PlayerVolume(&player,volume);
}

void setpan(int pan)
{
}

void flush(int t)
{
  PlayerFlush(&player,t);
}
  
int getoutputtime()
{
  return PlayerOutputTime(&player);
}

int getwrittentime()
{
  return PlayerWrittenTime(&player);
}

#define OUT_VER_U 0x11
static Out_Module out = {
  OUT_VER_U,
  (char *)YASAPI_TITLE_U16,
  65536+128,
  0, // hmainwindow
  0, // hdllinstance
  config,
  about,
  init,
  quit,
  open,
  close,
  write,
  canwrite,
  isplaying,
  pause,
  setvolume,
  setpan,
  flush,
  getoutputtime,
  getwrittentime
};

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
  return &out;
}


