/*
 * yasapi.h
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#ifndef __YASAPI_H__ // [
#define __YASAPI_H__

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <yasapi_param.h>

#ifdef __cplusplus // [
extern "C" {
#endif // ]

/*
 * major: 2
 * minor: 2
 * build: 0
 * beta: 2
 *
 * time: on
 * ring2: on
 * ring1: off
 * state_final: off
 * option_gapless: off
 * tail: off
 * api: off
 */

///////////////////////////////////////////////////////////////////////////////
typedef int32_t sample_t;
typedef enum _ECommand ECommand;
typedef enum _EEngine EEngine;
typedef const struct _IRingDitherVtbl IRingDitherVtbl;
typedef enum _ERingFormat ERingFormat;
typedef enum _ERingAlloc ERingAlloc;
typedef struct _Ring Ring;
typedef struct _EngineTime EngineTime;
typedef const struct _IEngineTimeVtbl IEngineTimeVtbl;
typedef const struct _IEngineVolumeVtbl IEngineVolumeVtbl;
typedef struct _Command Command;
typedef struct _EngineOptions EngineOptions;
typedef struct _EngineParam EngineParam;
typedef struct _EngineMonitor EngineMonitor;
typedef struct _Engine Engine;
typedef enum _EPlayerLimits EPlayerLimits;

///////////////////////////////////////////////////////////////////////////////
struct _IRingDitherVtbl {
  const char *pId;
  void (*Start)(Ring *pRing);
  sample_t (*Noise)(Ring *pRing, WORD nChannel);
};

extern IRingDitherVtbl gcIRingDitherNullVtbl;
extern IRingDitherVtbl gcIRingDitherRoundVtbl;
extern IRingDitherVtbl gcIRingDitherSimpleVtbl;

////////////////
enum _ERingFormat {
  eRingFormatBitsPerSample=32,
};

////////////////
enum _ERingAlloc {
  eRingCreate,
  eRingRealloc,
};

////////////////
struct _Ring {
  void (*Write)(Ring *pRing, const BYTE **ppSrc, DWORD dwSize);
  void (*Read)(Ring *pRing, BYTE **ppDst, DWORD dwSize);
  WAVEFORMATEXTENSIBLE wfx;

  struct {
    IRingDitherVtbl *lpVtbl;

    union {
      struct {
        sample_t noise;
      } round;

      struct {
        DWORD dwScaleNoise;
        sample_t noise;
      } simple;
    };
  } dither;

  Engine *pEngine;
  BYTE *_buf;
  BYTE *_rp;
  BYTE *_wp;
  BYTE *_mp;

  struct {
    WORD _uWritten;
    WORD _uAvailable;
  } frames;
};

int RingCreateFrames(Ring *pRing, Engine *pEngine, UINT32 uFrames);
void RingDestroy(Ring *pRing);

void RingReset(Ring *pRing);
int RingReallocFrames(Ring *pRing, Engine *pEngine, UINT32 uFrames);
int RingResetWrite(Ring *pRing, int bInitial);
WORD RingFramesMaximum(Ring *pRing);
int RingWrite(Ring *pRing, const BYTE *buf, DWORD dwSizeSource);
int RingReadFrames(Ring *pRing, BYTE *buf, WORD uFrames);

DWORD RingSourceSize(Ring *pRing);
DWORD RingSourceSizeAvailable(Ring *pRing);
DWORD RingSourceSizeWritten(Ring *pRing);
DWORD RingTargetSize(Ring *pRing);
DWORD RingTargetSizeWritten(Ring *pRing);

///////////////////////////////////////////////////////////////////////////////
enum _ECommand {
  eCommandNull,
  eCommandExit,
  eCommandStart,
  eCommandFlush,
  eCommandKill,
  ////////////////
  eCommandMax,
};

////////////////
enum _EEngine {
  // [
    eEngineNull,
    // eEngineBase [
      eEngineCriticalSection,
      eEngineTimerEvent,
      eEngineTimer,
      eEngineEventProducer,
      eEngineEventConsumer,
      eEngineThread,
      eEngineThreadRunning,
      eEngineBase=eEngineThreadRunning,
    // ]
    // eEngineInitialized [
      eEngineAudioClientInititial,
      eEngineRing,
      eEngineInitialized=eEngineRing,
    // ]
    // eEngineStarted [
      eEnginePauseDisconnected,
      eEngineDisconnected=eEnginePauseDisconnected,
      eEngineAudioClient,
      eEngineVolume,
      eEngineAudioClock,
      eEngineAudioRenderClient,
      eEngineStopped=eEngineAudioRenderClient,
      eEnginePauseConnected,
      eEngineAudioClientStarted,
      eEngineTimerStarted,
      eEngineStarted=eEngineTimerStarted,
    // ]
    eEngineDraining,
    eEngineEot,
   // ]
   eEngineMax,
};

////////////////
struct _EngineTime {
  IEngineTimeVtbl *lpVtbl;
  double ms;
};

////////////////
struct _IEngineTimeVtbl {
  const char *pId;
  int (*ClockCreate)(Engine *pEngine);
  void (*ClockDestroy)(Engine *pEngine);
  ////////////////
  void (*Inc)(Engine *pEngine);
  int (*Get)(Engine *pEngine);
  void (*Set)(Engine *pEngine, int ms);
};

extern IEngineTimeVtbl gcIEngineTimeNullVtbl;
extern IEngineTimeVtbl gcIEngineTimeClockVtbl;

////////////////
struct _IEngineVolumeVtbl {
  const char *pId;
  int bApplyGain;
  int (*Create)(Engine *pEngine);
  void (*Destroy)(Engine *pEngine);
  void (*Set)(Engine *pEngine);
};

extern IEngineVolumeVtbl gcIEngineVolumeNullVtbl;
extern IEngineVolumeVtbl gcIEngineVolumeSoftwareVtbl;
extern IEngineVolumeVtbl gcIEngineVolumeSimpleAudioVtbl;
extern IEngineVolumeVtbl gcIEngineVolumeAudioEndpointVtbl;

////////////////
struct _Command {
  ECommand eCommand;
  EEngine eArgument;
};

////////////////
int WfxCreate(WAVEFORMATEXTENSIBLE *pWfx, DWORD samplerate, WORD numchannels,
    WORD bitspersamp, int bSurround);

////////////////
struct _EngineParam {
  const char *pFriendlyName;
  int bDoubleBuffered;
  AUDCLNT_SHAREMODE eShareMode;
  DWORD dwStreamFlags;
  REFERENCE_TIME hnsBufferDuration;
  LARGE_INTEGER DueTime;
  LONG lPeriod;
  UINT32 uNumFramesDevice;
  UINT32 uNumFramesRing;
  DWORD dwChunkSizeSource;
  WORD uFramesPlaybackBegin;
  double xFramesPlaybackBegin;
  int nLastPause;
  EngineTime time;
  IEngineVolumeVtbl *pIEngineVolumeVtbl;
  float fLevel;
};

struct _EngineOptions {
  OptionsCommon common;
  OptionsDevice device;
};

struct _EngineMonitor {
  double xPadding;
  double xWritten;
  double xStart;

};

struct _Engine {
  EEngine eState;
  Player *pPlayer;
  CRITICAL_SECTION cs;

  struct {
    UINT32 uNumTracks;
    UINT32 uNumFramesMax;
    UINT32 uNumDropouts;
  } stats;

  struct {
    HANDLE hEvent;
    HANDLE hTimer;
  } timer;

  struct {
    Command *pCommand;
    HANDLE hProducer;     // When we're the producer we're waiting
                          // on this event.
    HANDLE hConsumer;     // When we're the consumer we're waiting
                          // on this event.
    HANDLE hThread;
  } thread;

  IMMDevice *pDevice;

  EngineOptions options;
  EngineParam param;
  EngineMonitor monitor;

  struct {
    WAVEFORMATEXTENSIBLE source;
    WAVEFORMATEXTENSIBLE target;
  } wfx;

  Ring ring;

  union {
    void *pVolumeVoid;
    ISimpleAudioVolume *pSimpleAudioVolume;
    IAudioEndpointVolume *pAudioEndpointVolume;
  };

  IAudioClient *pAudioClient;
  IAudioClock *pAudioClock;
  IAudioRenderClient *pAudioRenderClient;
};

int EngineCreate(Engine *pEngine, Player *pPlayer, int samplerate,
    int numchannels, int bitspersamp);
int EngineCreateAudioClient(Engine *pEngine, const EEngine eState);
void EngineKill(Engine *pEngine, const EEngine eState);
void EngineDestroy(Engine *pEngine);
void EngineClose(Engine *pEngine);

int EngineNumFramesPadding(Engine *pEngine, UINT32 *puNumFramesPadding);
int EnginePlayingStart(Engine *pEngine);
int EnginePause(Engine *pEngine, int nPause);
int EngineIsPlaying(Engine *pEngine);
int EngineAvailable(Engine *pEngine);
int EngineOutputTime(Engine *pEngine);
int EngineWrittenTime(Engine *pEngine);
int EngineWrite(Engine *pEngine, BYTE *buf, DWORD dwChunkSize);

void EngineSend(Engine *pEngine, const ECommand eCommand, EEngine eState,
    int bLock);
int EngineStart(Engine *pEngine, int bAudioClient, int bTimer);
int EnginePlay(Engine *pEngine);
void EngineFlush(Engine *pEngine);
void EngineVolume(Engine *pEngine, Player *pPlayer);

DWORD WINAPI EngineThreadProc(LPVOID lpParameter);

const char *EngineStateEx(EEngine eState);
const char *EngineState(Engine *pEngine);
const char *EngineCommandEx(ECommand eCommand);
const char *EngineCommand(Engine *pEngine);

///////////////////////////////////////////////////////////////////////////////
enum _EPlayerLimits {
  ePlayerVolumeMax=255,
};

struct _Player {
  Out_Module *pOut;
  wchar_t ini[MAX_PATH];

  struct {
    int nVolume;
  } param;

  struct {
    UINT32 uNumTracks;
    UINT32 uNumDropouts;
    UINT32 uMsLatencyMax;
  } stats;

  Wasapi wasapi;
  Engine engine;
  Gtk gtk;
};

int PlayerCreate(Player *pPlayer, Out_Module *pOut);
void PlayerDestroy(Player *pPlayer);

int PlayerPropertiesPath(Player *pPlayer);

int PlayerOpen(Player *pPlayer, int samplerate, int numchannels,
    int bitspersamp);
void PlayerClose(Player *pPlayer);

Wasapi *PlayerWasapi(Player *pPlayer);
void PlayerFlush(Player *pPlayer, int ms);
int PlayerPause(Player *pPlayer, int nPause);
int PlayerIsPlaying(Player *pPlayer);
int PlayerAvailable(Player *pPlayer);
int PlayerOutputTime(Player *pPlayer);
int PlayerWrittenTime(Player *pPlayer);
int PlayerWrite(Player *pPlayer, BYTE *buf, int len);
void PlayerVolume(Player *pPlayer, int nVolume);

int PlayerShow(Player *pPlayer, Dialog *pDialog, HWND hWndParent);

///////////////////////////////////////////////////////////////////////////////
const char *Bitdepth(EBitdepth eBitdepth);

#ifdef __cplusplus // [
}
#endif // ]
#endif // __YASAPI_H__ ]
