/*
 * yasapi_config_combo_box.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <yasapi_config.h>

static void ConfigComboBoxSet(ConfigWidget *pWidget)
{
  Config *pConfig=&pWidget->pDialog->config;
  ClassCell *pCell=pWidget->pCell;
  int *value=(int *)pCell->nav(pWidget->pDialog,pCell->offs);
  int bOptionsDevice=NavigateOptionsDeviceConfig==pCell->nav;
  int bMode=bOptionsDevice
      &&pCell->offs==offsetof(OptionsDevice,eMode);
  int bStrategy=bOptionsDevice
      &&pCell->offs==offsetof(OptionsDevice,eStrategy);
  ConfigComboBoxMap *map;
  ConfigWidget *pWidgetSideEffect;
  int i;

  for (map=pCell->data,i=0;map->label;++map,++i) {
    if (map->value==*value) {
      gtk_combo_box_set_active (GTK_COMBO_BOX (pWidget->widget), i);

      if (bMode) {
        ////////
        pWidgetSideEffect=&pConfig->wasapi.auto_convert_pcm;

        if (pWidgetSideEffect->widget)
          gtk_widget_set_sensitive (pWidgetSideEffect->widget,!map->value);

        ////////
        pWidgetSideEffect=&pConfig->wasapi.src_default_quality;

        if (pWidgetSideEffect->widget)
          gtk_widget_set_sensitive (pWidgetSideEffect->widget,!map->value);

      }
      else if (bStrategy) {
        ////////
        pWidgetSideEffect=&pConfig->wasapi.timer_periodic;

        if (pWidgetSideEffect->widget)
          gtk_widget_set_sensitive (pWidgetSideEffect->widget,!map->value);
      }

      return;
    }
  }
}

static void ConfigComboBoxGet(ConfigWidget *pWidget)
{
  ClassCell *pCell=pWidget->pCell;
  int active = gtk_combo_box_get_active (GTK_COMBO_BOX (pWidget->widget));
  ConfigComboBoxMap *map = (ConfigComboBoxMap *)pCell->data + active;
  int bOptionsDevice=NavigateOptionsDeviceConfig==pCell->nav;
  int bMode=bOptionsDevice
      &&pCell->offs==offsetof(OptionsDevice,eMode);
  OptionsDevice *pOptions;
  int *value;

  if (bMode) {
    pOptions=(OptionsDevice *)pCell->nav(pWidget->pDialog,0);
    pOptions->eMode=map->value;
    pOptions->ePeriod=map->value?ePeriodMinimum:ePeriodDefault;
  }
  else {
    value=(int *)pCell->nav(pWidget->pDialog,pCell->offs);
    *value = map->value;
  }
}

static void yasapi_config_combo_box_changed (GtkComboBox *widget,
    gpointer user_data)
{
  Dialog *pDialog=user_data;
  ConfigWidget *pWidget=(ConfigWidget *)g_object_get_data(G_OBJECT(widget),
      CONFIG_WIDGET);

  pWidget->lpVtbl->get(pWidget);
  pDialog->lpVtbl->set(pDialog);
}

static GtkWidget *ConfigComboBoxCreate(ClassCell *pCell, Dialog *pDialog,
    Expand *pExpand)
{
  static IConfigWidgetVtbl vtbl={
    .set=ConfigComboBoxSet,
    .get=ConfigComboBoxGet,
  };

  GtkWidget *widget = gtk_combo_box_text_new();
  ConfigComboBoxMap *map;

  for (map=pCell->data;map->label;++map) {
    gtk_combo_box_text_append (GTK_COMBO_BOX_TEXT (widget), NULL, 
        map->label);
  }

  ConfigWidgetCreate(pCell, pDialog, &vtbl, widget);

  g_signal_connect (widget, "changed",
      G_CALLBACK (yasapi_config_combo_box_changed), pDialog);

  return widget;
}

IClassCellVtbl gcIClassComboBoxVtbl={
  .create=ConfigComboBoxCreate,
};
