/*
 * yasapi_engine_thread.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int EngineRender(Engine *pEngine, UINT32 uFrames)
{
  int code=-1;
  WORD nBlockAlign=pEngine->wfx.target.Format.nBlockAlign;
  int bSilence=0;
  DWORD dwSizeRead=nBlockAlign*uFrames;
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  Ring *pRing=&pEngine->ring;
  DWORD dwSize=RingTargetSize(pRing);
  DWORD dwWritten=RingTargetSizeWritten(pRing);
  IAudioRenderClient *pAudioRenderClient=pEngine->pAudioRenderClient;
  BYTE *pData;
  HRESULT hr;

  if (pEngine->eState<eEngineAudioRenderClient) {
    DVMESSAGE("internal %s",EngineState(pEngine));
    goto exit;
  }

  pEngine->monitor.xWritten=(double)dwWritten/dwSize;
  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.written);

  if (dwSizeRead<=dwWritten) {
    //DVWRITELN("*** available %s ***",EngineState(pEngine));
  }
  else if (pEngine->eState<eEngineDraining) {
    bSilence=1;

    YASAPI_UINT32_INC(pEngine->stats.uNumDropouts);
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.num_dropouts);
    ++pEngine->pPlayer->stats.uNumDropouts;
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.player.num_dropouts);
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pAudioRenderClient->lpVtbl->GetBuffer(pAudioRenderClient,
    uFrames,                // [in]  UINT32 NumFramesRequested,
    &pData                  // [out] BYTE   **ppData
  );

  if (FAILED(hr)) {
    switch (hr) {
    DVCMESSAGE("IAudioRenderClient::GetBuffer: uNumFrams=%u (%u)",
        AUDCLNT_E_BUFFER_ERROR,uFrames,pEngine->param.uNumFramesDevice);
    DVCMESSAGE("IAudioRenderClient::GetBuffer: uNumFrams=%u (%u)",
        AUDCLNT_E_BUFFER_TOO_LARGE,uFrames,pEngine->param.uNumFramesDevice);
    DVCMESSAGE("IAudioRenderClient::GetBuffer: uNumFrams=%u (%u)",
        AUDCLNT_E_BUFFER_SIZE_ERROR,uFrames,pEngine->param.uNumFramesDevice);
    DCMESSAGE("IAudioRenderClient::GetBuffer",AUDCLNT_E_OUT_OF_ORDER);
    DCMESSAGE("IAudioRenderClient::GetBuffer",
        AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioRenderClient::GetBuffer",
        AUDCLNT_E_BUFFER_OPERATION_PENDING);
    DCMESSAGE("IAudioRenderClient::GetBuffer",
        AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioRenderClient::GetBuffer",E_POINTER);
    DDMESSAGE("IAudioRenderClient::GetBuffer",hr);
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bSilence)
    ZeroMemory(pData,dwSizeRead);
  else if (dwSizeRead<=dwWritten) {
    if (RingReadFrames(&pEngine->ring,pData,uFrames)<0) {
      DMESSAGE("reading ring");
      goto eread;
    }
  }
  else if (0ul<dwWritten) {
    if (RingReadFrames(&pEngine->ring,pData,dwWritten/nBlockAlign)<0) {
      DMESSAGE("reading ring");
      goto eread;
    }

    ZeroMemory(pData+dwWritten,dwSizeRead-dwWritten);
  }
  else
    ZeroMemory(pData,dwSizeRead);

  /////////////////////////////////////////////////////////////////////////////
  code=0;
eread:
  hr=pAudioRenderClient->lpVtbl->ReleaseBuffer(pAudioRenderClient,
    uFrames,                // [in] UINT32 NumFramesWritten,
    0ul                     // [in] DWORD  dwFlags
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioRenderClient::ReleaseBuffer",AUDCLNT_E_INVALID_SIZE);
    DCMESSAGE("IAudioRenderClient::ReleaseBuffer",
        AUDCLNT_E_BUFFER_SIZE_ERROR);
    DCMESSAGE("IAudioRenderClient::ReleaseBuffer",AUDCLNT_E_OUT_OF_ORDER);
    DCMESSAGE("IAudioRenderClient::ReleaseBuffer",
        AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioRenderClient::ReleaseBuffer",
        AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioRenderClient::ReleaseBuffer",E_INVALIDARG);
    DDMESSAGE("IAudioRenderClient::ReleaseBuffer",hr);
    }

    // nowhere to go ...
  }
exit:
  return code;
}

int EngineStart(Engine *pEngine, int bAudioClient, int bTimer)
{
  int code=-1;
  BOOL bCode;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (bAudioClient) {
    hr=pEngine->pAudioClient->lpVtbl->Start(pEngine->pAudioClient);

    if (FAILED(hr)) {
      switch (hr) {
      DCMESSAGE("IAudioClient::Start",AUDCLNT_E_NOT_INITIALIZED);
      DCMESSAGE("IAudioClient::Start",AUDCLNT_E_NOT_STOPPED);
      DCMESSAGE("IAudioClient::Start",AUDCLNT_E_EVENTHANDLE_NOT_SET);
      DCMESSAGE("IAudioClient::Start",AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE("IAudioClient::Start",AUDCLNT_E_SERVICE_NOT_RUNNING);
      DDMESSAGE("IAudioClient::Start",hr);
      }

      goto exit;
    }

    if (pEngine->eState<eEnginePauseConnected) {
      ++pEngine->eState;
      DASSERT(eEnginePauseConnected==pEngine->eState,exit);
    }

    ++pEngine->eState;
    DASSERT(eEngineAudioClientStarted==pEngine->eState,exit);
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bTimer) {
    if (!pEngine->options.device.eStrategy) {

      bCode=SetWaitableTimer(
        pEngine->timer.hTimer,
                    // _In_           HANDLE           hTimer,
        &pEngine->param.DueTime,
                    // _In_     const LARGE_INTEGER    *pDueTime,
        pEngine->param.lPeriod,
                    // _In_           LONG             lPeriod,
        NULL,       // _In_opt_       PTIMERAPCROUTINE pfnCompletionRoutine,
        NULL,       // _In_opt_       LPVOID           lpArgToCompletionRoutine,
        FALSE       // _In_           BOOL             fResume
      );

      if (!bCode) {
        DMESSAGE("starting waitable timer");
        goto exit;
      }
    }
  }

  if (bAudioClient) {
    ++pEngine->eState;
    DASSERT(eEngineTimerStarted==pEngine->eState,exit);
  }

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

int EnginePlayingStart(Engine *pEngine)
{
  int code=-1;
  const EEngine eState=pEngine->eState;

  if (EngineStart(pEngine,1,!pEngine->options.device.eStrategy)<0) {
    DMESSAGE("starting");
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

static int EngineConnect(Engine *pEngine)
{
  const EEngine eState=pEngine->eState;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (!pEngine->pAudioClient&&EngineCreateAudioClient(pEngine,eState+1)<0) {
    DMESSAGE("creating audio client");
    EngineKill(pEngine,eState);
    goto exit;
  }

  if (pEngine->eState<eEnginePauseDisconnected) {
    ++pEngine->eState;
    DWRITELN(EngineState(pEngine));
    DASSERT(eEnginePauseDisconnected==pEngine->eState,exit);
  }

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT(eEngineAudioClient==pEngine->eState,exit);

	if (pEngine->param.pIEngineVolumeVtbl->Create(pEngine)<0) {
    EngineKill(pEngine,eState);
		goto exit;
	}

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT(eEngineVolume==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->param.time.lpVtbl->ClockCreate(pEngine)<0) {
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT(eEngineAudioClock==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  hr=pEngine->pAudioClient->lpVtbl->GetService(pEngine->pAudioClient,
    &IID_IAudioRenderClient,    // [in]  REFIID riid,
    (void **)&pEngine->pAudioRenderClient
                                // [out] void   **ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetService",E_POINTER);
    DCMESSAGE("IAudioClient::GetService",E_NOINTERFACE);
    DCMESSAGE("IAudioClient::GetService",AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE("IAudioClient::GetService",AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    DCMESSAGE("IAudioClient::GetService",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetService",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClient::GetService",hr);
    }

    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT(eEngineAudioRenderClient==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
exit:
  return eState<pEngine->eState?0:-1;
}

int EngineNumFramesPadding(Engine *pEngine, UINT32 *puNumFramesPadding)
{
  int code=-1;
  HRESULT hr;

  if (pEngine->eState<eEngineAudioClient)
    goto exit;

  hr=pEngine->pAudioClient->lpVtbl->GetCurrentPadding(pEngine->pAudioClient,
    puNumFramesPadding    // [out] UINT32 *pNumPaddingFrames
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetCurrentPadding",AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE("IAudioClient::GetCurrentPadding",
        AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetCurrentPadding",
        AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioClient::GetCurrentPadding",E_POINTER);
    DDMESSAGE("IAudioClient::GetCurrentPadding",hr);
    }

    goto exit;
  }

  code=0;
exit:
  return code;
}

int EnginePlay(Engine *pEngine)
{
  int code=-1;
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  UINT32 uNumFramesPadding;
  UINT32 uNumFrames;

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->eState<eEngineAudioClient&&EngineConnect(pEngine)<0) {
    DMESSAGE("connecting");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->param.bDoubleBuffered)
    uNumFrames=pEngine->param.uNumFramesDevice;
  else {
    if (!pEngine->pAudioClient) {
      DMESSAGE("attempt to access invalidated pAudioClient");
      goto exit;
    }

    if (EngineNumFramesPadding(pEngine,&uNumFramesPadding)<0) {
      DMESSAGE("getting padding");
      goto exit;
    }

    pEngine->monitor.xPadding
        =(double)uNumFramesPadding/pEngine->param.uNumFramesDevice;
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.padding);

    if (0u==uNumFramesPadding) {
      uNumFrames=pEngine->param.uNumFramesDevice;
    }
    else if (pEngine->param.uNumFramesDevice<uNumFramesPadding) {
      // Should never happen, but unfortunately sometimes it does.
      uNumFrames=0u;
    }
    else
      uNumFrames=pEngine->param.uNumFramesDevice-uNumFramesPadding;
  }

  if (EngineRender(pEngine,uNumFrames)<0) {
    DMESSAGE("rendering");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->eState<eEngineStarted) {
    if (EnginePlayingStart(pEngine)<0) {
      DMESSAGE("starting engine");
      goto exit;
    }

    DWRITELN("*** start playing ***");
  }
  else if (EngineStart(pEngine,0,1)<0) {
    DMESSAGE("start timer");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

void EngineFlush(Engine *pEngine)
{
  HRESULT hr;

  if (pEngine->eState<eEngineStarted)
    goto exit;

  /////////////////////////////////////////////////////////////////////////////
  EngineKill(pEngine,eEngineStopped);

  /////////////////////////////////////////////////////////////////////////////
  hr=pEngine->pAudioClient->lpVtbl->Reset(pEngine->pAudioClient);

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::Reset",AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE("IAudioClient::Reset",AUDCLNT_E_NOT_STOPPED);
    DCMESSAGE("IAudioClient::Reset",AUDCLNT_E_BUFFER_OPERATION_PENDING);
    DCMESSAGE("IAudioClient::Reset",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClient::Reset",hr);
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  pEngine->param.time.lpVtbl->Set(pEngine,0);

  /////////////////////////////////////////////////////////////////////////////
  RingReset(&pEngine->ring);
exit:
  return;
}
