/*
 * yasapi_engine_volume.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID_TIME(id) \
  YASAPI_PASTE3(EngineTime,YASAPI_ID,id)
#define YASAPI_ID_ITIME_VTBL(id) \
  YASAPI_PASTE3(gcIEngineTime,id,Vtbl)

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID Null

static int YASAPI_ID_TIME(ClockCreate)(Engine *pEngine)
{
  return 0;
}

static void YASAPI_ID_TIME(ClockDestroy)(Engine *pEngine)
{
}

static void YASAPI_ID_TIME(Inc)(Engine *pEngine)
{
}

static int YASAPI_ID_TIME(Get)(Engine *pEngine)
{
  return 0;
}

static void YASAPI_ID_TIME(Set)(Engine *pEngine, int ms)
{
  pEngine->param.time.ms=ms;
}

IEngineTimeVtbl YASAPI_ID_ITIME_VTBL(YASAPI_ID)={
  .pId=YASAPI_STRA(YASAPI_ID),
  .ClockCreate=YASAPI_ID_TIME(ClockCreate),
  .ClockDestroy=YASAPI_ID_TIME(ClockDestroy),
  .Inc=YASAPI_ID_TIME(Inc),
  .Get=YASAPI_ID_TIME(Get),
  .Set=YASAPI_ID_TIME(Set),
};

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID Clock

static int YASAPI_ID_TIME(ClockCreate)(Engine *pEngine)
{
  HRESULT hr;

  hr=pEngine->pAudioClient->lpVtbl->GetService(pEngine->pAudioClient,
    &IID_IAudioClock,         // [in]  REFIID riid,
    (void **)&pEngine->pAudioClock
                              // [out] void   **ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",E_POINTER);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",E_NOINTERFACE);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetService(IAudioClock)",
        AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClient::GetService(IAudioClock)",hr);
    }

    goto exit;
  }

  return 0;
exit:
  return -1;
}

static void YASAPI_ID_TIME(ClockDestroy)(Engine *pEngine)
{
  pEngine->pAudioClock->lpVtbl->Release(pEngine->pAudioClock);
  pEngine->pAudioClock=NULL;
}

static double _EngineTimeClockGet(Engine *pEngine)
{
  double ms=0.0;
  UINT64 u64Frequency;
  UINT64 u64Position;
  HRESULT hr;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs);

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
  case eEngineEot:
    goto exit;
  default:
    if (pEngine->eState<eEngineAudioClock)
      goto exit;

    break;
  }

  hr=pEngine->pAudioClock->lpVtbl->GetFrequency(pEngine->pAudioClock,
    &u64Frequency     // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetFrequency",E_POINTER);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetFrequency",hr);
    }

    goto exit;
  }

  hr=pEngine->pAudioClock->lpVtbl->GetPosition(pEngine->pAudioClock,
    &u64Position,     // [out] UINT64 *pu64Position
    NULL              // UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetPosition",E_POINTER);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetPosition",hr);
    }

    goto exit;
  }

  ms=1000.0*u64Position/u64Frequency;
exit:
  LeaveCriticalSection(&pEngine->cs);
ebase:
  return ms;
}

static void YASAPI_ID_TIME(Inc)(Engine *pEngine)
{
  pEngine->param.time.ms+=_EngineTimeClockGet(pEngine);
}

static int YASAPI_ID_TIME(Get)(Engine *pEngine)
{
  return _EngineTimeClockGet(pEngine)-pEngine->param.time.ms+0.5;
}

static void YASAPI_ID_TIME(Set)(Engine *pEngine, int ms)
{
  pEngine->param.time.ms=ms;
}

IEngineTimeVtbl YASAPI_ID_ITIME_VTBL(YASAPI_ID)={
  .pId=YASAPI_STRA(YASAPI_ID),
  .ClockCreate=YASAPI_ID_TIME(ClockCreate),
  .ClockDestroy=YASAPI_ID_TIME(ClockDestroy),
  .Inc=YASAPI_ID_TIME(Inc),
  .Get=YASAPI_ID_TIME(Get),
  .Set=YASAPI_ID_TIME(Set),
};
