/*
 * yasapi_engine_volume.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID_VOLUME(id) \
  YASAPI_PASTE3(EngineVolume,YASAPI_ID,id)
#define YASAPI_ID_IVOLUME_VTBL(id) \
  YASAPI_PASTE3(gcIEngineVolume,id,Vtbl)

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID Null

static int YASAPI_ID_VOLUME(Create)(Engine *pEngine)
{
  Gtk *pGtk=&pEngine->pPlayer->gtk;

  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.volume);

  return 0;
}

static void YASAPI_ID_VOLUME(Destroy)(Engine *pEngine)
{
}

static void YASAPI_ID_VOLUME(Set)(Engine *pEngine)
{
}

IEngineVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="Off",
  .bApplyGain=0,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID Software

static int YASAPI_ID_VOLUME(Create)(Engine *pEngine)
{
  Gtk *pGtk=&pEngine->pPlayer->gtk;

  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.volume);

  return 0;
}

static void YASAPI_ID_VOLUME(Destroy)(Engine *pEngine)
{
}

static void YASAPI_ID_VOLUME(Set)(Engine *pEngine)
{
}

IEngineVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="YASAPI/NT",
  .bApplyGain=1,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};

static IEngineVolumeVtbl YASAPI_ID_IVOLUME_VTBL(SoftwareAudioSimple)={
  .pId="YASAPI/NT (Fallback for ISimpleAudioVolume)",
  .bApplyGain=1,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};

static IEngineVolumeVtbl YASAPI_ID_IVOLUME_VTBL(SoftwareAudioEndpoint)={
  .pId="YASAPI/NT (Fallback for IAudioEndpointVolume)",
  .bApplyGain=1,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID SimpleAudio

static void YASAPI_ID_VOLUME(Set)(Engine *pEngine);

static int YASAPI_ID_VOLUME(Create)(Engine *pEngine)
{
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  HRESULT hr;

  hr=pEngine->pAudioClient->lpVtbl->GetService(pEngine->pAudioClient,
    &IID_ISimpleAudioVolume,    // [in]  REFIID riid,
    (void **)&pEngine->pSimpleAudioVolume
                                // [out] void   **ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetService(ISimpleAudioVolume)",E_POINTER);
    DCMESSAGE("IAudioClient::GetService(ISimpleAudioVolume)",E_NOINTERFACE);
    DCMESSAGE("IAudioClient::GetService(ISimpleAudioVolume)",
        AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE("IAudioClient::GetService(ISimpleAudioVolume)",
        AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    DCMESSAGE("IAudioClient::GetService(ISimpleAudioVolume)",
        AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetService(ISimpleAudioVolume)",
        AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClient::GetService(ISimpleAudioVolume)",hr);
    }

    goto eactivate;
  }

  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.volume);
  YASAPI_ID_VOLUME(Set)(pEngine);

  return 0;
eactivate:
  pEngine->param.pIEngineVolumeVtbl
      =&YASAPI_ID_IVOLUME_VTBL(SoftwareAudioSimple);

  return pEngine->param.pIEngineVolumeVtbl->Create(pEngine);
}

static void YASAPI_ID_VOLUME(Destroy)(Engine *pEngine)
{
  pEngine->pSimpleAudioVolume->lpVtbl->Release(pEngine->pSimpleAudioVolume);
}

static void YASAPI_ID_VOLUME(Set)(Engine *pEngine)
{
  ISimpleAudioVolume *pSimpleAudioVolume=pEngine->pSimpleAudioVolume;
  int nVolume=pEngine->pPlayer->param.nVolume;
  float fLevel;

  if (nVolume<0||ePlayerVolumeMax<nVolume)
    return;

  fLevel=(double)pEngine->pPlayer->param.nVolume/ePlayerVolumeMax;

  pSimpleAudioVolume->lpVtbl->SetMasterVolume(pSimpleAudioVolume,
    fLevel,       // [in] float   fLevel,
    NULL          // [in] LPCGUID EventContext
  );
}

IEngineVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="ISimpleAudioVolume",
  .bApplyGain=0,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_ID) // [
#undef YASAPI_ID
#endif // ]
#define YASAPI_ID AudioEndpoint

static void YASAPI_ID_VOLUME(Set)(Engine *pEngine);

static int YASAPI_ID_VOLUME(Create)(Engine *pEngine)
{
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  DWORD dwHardwareSupportMask=0ul;
  IAudioEndpointVolume *pAudioEndpointVolume;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  hr=pEngine->pDevice->lpVtbl->Activate(pEngine->pDevice,
    &IID_IAudioEndpointVolume,      // [in]  REFIID      iid,
    CLSCTX_ALL,                     // [in]  DWORD       dwClsCtx,
    NULL,                           // [in]  PROPVARIANT *pActivationParams,
    (void **)&pAudioEndpointVolume  // [out] void        **ppInterface
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IMMDevice::Activate(IAudioEndpointVolume)",E_NOINTERFACE);
    DCMESSAGE("IMMDevice::Activate(IAudioEndpointVolume)",E_POINTER);
    DCMESSAGE("IMMDevice::Activate(IAudioEndpointVolume)",E_INVALIDARG);
    DCMESSAGE("IMMDevice::Activate(IAudioEndpointVolume)",E_OUTOFMEMORY);
    DCMESSAGE("IMMDevice::Activate(IAudioEndpointVolume)",
        AUDCLNT_E_DEVICE_INVALIDATED);
    DDMESSAGE("IMMDevice::Activate(IAudioEndpointVolume)",hr);
    }

    goto eactivate;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pAudioEndpointVolume->lpVtbl->QueryHardwareSupport(
    pAudioEndpointVolume,
    &dwHardwareSupportMask          // [out] DWORD *pdwHardwareSupportMask
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioEndpointVolume::QueryHardwareSupport",E_POINTER);
    DDMESSAGE("IAudioEndpointVolume::QueryHardwareSupport",hr);
    }

    goto ehardware;
  }
  else if (!(ENDPOINT_HARDWARE_SUPPORT_VOLUME&dwHardwareSupportMask))
    goto ehardware;

  /////////////////////////////////////////////////////////////////////////////
  hr=pAudioEndpointVolume->lpVtbl->GetMasterVolumeLevelScalar(
    pAudioEndpointVolume,
    &pEngine->param.fLevel         // [out] float *pfLevel
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioEndpointVolume::GetMasterVolumeLevelScalar",E_POINTER);
    DDMESSAGE("IAudioEndpointVolume::GetMasterVolumeLevelScalar",hr);
    }

    goto elevel;
  }

  pEngine->pAudioEndpointVolume=pAudioEndpointVolume;
  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.volume);
  YASAPI_ID_VOLUME(Set)(pEngine);

  return 0;
//error:
  pAudioEndpointVolume->lpVtbl->SetMasterVolumeLevelScalar(
      pAudioEndpointVolume,pEngine->param.fLevel,NULL);
elevel:
ehardware:
  pAudioEndpointVolume->lpVtbl->Release(pAudioEndpointVolume);
eactivate:
  pEngine->param.pIEngineVolumeVtbl
      =&YASAPI_ID_IVOLUME_VTBL(SoftwareAudioEndpoint);

  return pEngine->param.pIEngineVolumeVtbl->Create(pEngine);
}

static void YASAPI_ID_VOLUME(Destroy)(Engine *pEngine)
{
  IAudioEndpointVolume *pAudioEndpointVolume=pEngine->pAudioEndpointVolume;

  pAudioEndpointVolume->lpVtbl->SetMasterVolumeLevelScalar(
      pAudioEndpointVolume,pEngine->param.fLevel,NULL);
  pAudioEndpointVolume->lpVtbl->Release(pAudioEndpointVolume);
}

static void YASAPI_ID_VOLUME(Set)(Engine *pEngine)
{
  IAudioEndpointVolume *pAudioEndpointVolume=pEngine->pAudioEndpointVolume;
  int nVolume=pEngine->pPlayer->param.nVolume;
  HRESULT hr;
  float fLevel;

  if (nVolume<0||ePlayerVolumeMax<nVolume)
    return;

  fLevel=(double)pEngine->pPlayer->param.nVolume/ePlayerVolumeMax;

  hr=pAudioEndpointVolume->lpVtbl->SetMasterVolumeLevelScalar(
    pAudioEndpointVolume,
    fLevel,     // [in] float   fLevel,
    NULL        // [in] LPCGUID pguidEventContext
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioEndpointVolume::SetMasterVolumeLevelScalar",
        E_INVALIDARG);
    DCMESSAGE("IAudioEndpointVolume::SetMasterVolumeLevelScalar",
        E_OUTOFMEMORY);
    DDMESSAGE("IAudioEndpointVolume::SetMasterVolumeLevelScalar",hr);
    }
  }
}

IEngineVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="IAudioEndpointVolume",
  .bApplyGain=0,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};
