/*
 * yasapi_ring2.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
static void RingWriteSamples(Ring *pRing, const BYTE **ppSrc, DWORD dwSize)
{
  BYTE *mp=pRing->_wp+dwSize;
  WAVEFORMATEX *pSource=&pRing->pEngine->wfx.source.Format;
  WAVEFORMATEX *pFormat=&pRing->wfx.Format;
  sample_t sample;
  BYTE *pSampeWrite=(BYTE *)&sample+((sizeof sample)
      -(pSource->wBitsPerSample>>3));
  const BYTE *pSampeRead=(const BYTE *)&sample+((sizeof sample)
      -(pFormat->wBitsPerSample>>3));

  while (pRing->_wp<mp) {
    sample=0;

    memcpy(pSampeWrite,*ppSrc,pSource->wBitsPerSample>>3);
    *ppSrc+=pSource->wBitsPerSample>>3;

    memcpy(pRing->_wp,pSampeRead,pFormat->wBitsPerSample>>3);
    pRing->_wp+=pFormat->wBitsPerSample>>3;
  }

  if (pRing->_mp<=pRing->_wp)
    pRing->_wp=pRing->_buf;
}

static void RingWriteSamplesDuplicate(Ring *pRing, const BYTE **ppSrc,
    DWORD dwSize)
{
  BYTE *mp=pRing->_wp+dwSize;
  WAVEFORMATEX *pSource=&pRing->pEngine->wfx.source.Format;
  WAVEFORMATEX *pFormat=&pRing->wfx.Format;
  sample_t sample;
  BYTE *pSampeWrite=(BYTE *)&sample+((sizeof sample)
      -(pSource->wBitsPerSample>>3));
  const BYTE *pSampeRead=(const BYTE *)&sample+((sizeof sample)
      -(pFormat->wBitsPerSample>>3));

  while (pRing->_wp<mp) {
    sample=0;

    memcpy(pSampeWrite,*ppSrc,pSource->wBitsPerSample>>3);
    *ppSrc+=pSource->wBitsPerSample>>3;

    memcpy(pRing->_wp,pSampeRead,pFormat->wBitsPerSample>>3);
    pRing->_wp+=pFormat->wBitsPerSample>>3;
    memcpy(pRing->_wp,pSampeRead,pFormat->wBitsPerSample>>3);
    pRing->_wp+=pFormat->wBitsPerSample>>3;
  }

  if (pRing->_mp<=pRing->_wp)
    pRing->_wp=pRing->_buf;
}

static void RingReadSamples(Ring *pRing, BYTE **ppDst, DWORD dwSize)
{
	static double xSqrt2Inverse=-1.0;

  const BYTE *mp=pRing->_rp+dwSize;
  WAVEFORMATEX *pFormat=&pRing->wfx.Format;
  WAVEFORMATEX *pSource=&pRing->pEngine->wfx.source.Format;
  WAVEFORMATEX *pTarget=&pRing->pEngine->wfx.target.Format;
  int bApplyGain=pRing->pEngine->param.pIEngineVolumeVtbl->bApplyGain;
  int nVolume=pRing->pEngine->pPlayer->param.nVolume;
  EMono2Stereo eMono2Stereo=pRing->pEngine->options.common.eMono2Stereo;
  int bDither;
  WORD nChannel;
  sample_t sample;
  BYTE *pSampleWrite=(BYTE *)&sample+((sizeof sample)
      -(pFormat->wBitsPerSample>>3));
  const BYTE *pSampleRead=(const BYTE *)&sample+((sizeof sample)
      -(pTarget->wBitsPerSample>>3));

  bApplyGain=bApplyGain&&(0<=nVolume);
  bApplyGain=bApplyGain&&nVolume<ePlayerVolumeMax;
  pRing->dither.lpVtbl->Start(pRing);

  while (pRing->_rp<mp) {
    for (nChannel=0u;nChannel<pFormat->nChannels;++nChannel) {
      bDither=0;
      sample=0;

      memcpy(pSampleWrite,pRing->_rp,pFormat->wBitsPerSample>>3);
      pRing->_rp+=pFormat->wBitsPerSample>>3;

      if (bApplyGain) {
        bDither=1;
        sample=MulDiv(sample,nVolume,ePlayerVolumeMax);
      }

      if (pSource->nChannels!=pTarget->nChannels) {
        switch (eMono2Stereo) {
        case eMono2StereoHalf:
          bDither=1;
          sample>>=1;
          break;
        case eMono2StereoSqrt2Inverse:
          if (xSqrt2Inverse<0.0)
            xSqrt2Inverse=1.0/sqrt(2.0);

          bDither=1;
          sample=xSqrt2Inverse*sample+0.5;
          break;
        case eMono2StereoPlain:
        default:
          break;
        }
      }

      if (pTarget->wBitsPerSample<pSource->wBitsPerSample)
        bDither=1;

      if (bDither)
        sample+=pRing->dither.lpVtbl->Noise(pRing,nChannel);

      memcpy(*ppDst,pSampleRead,pTarget->wBitsPerSample>>3);
      *ppDst+=pTarget->wBitsPerSample>>3;
    }
  }

  if (pRing->_mp<=pRing->_rp)
    pRing->_rp=pRing->_buf;
}

///////////////////////////////////////////////////////////////////////////////
int RingCreateFrames(Ring *pRing, Engine *pEngine, UINT32 uFrames)
{
  int code=-1;
  EMono2Stereo eMono2Stereo=pEngine->options.common.eMono2Stereo;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WAVEFORMATEX *pFormat=&pRing->wfx.Format;
  DWORD dwRingSize;

  /////////////////////////////////////////////////////////////////////////////
  pRing->pEngine=pEngine;

DVWRITELN("uFrames: %u =======================================",uFrames);
  /////////////////////////////////////////////////////////////////////////////
  if (eMono2Stereo&&pSource->nChannels!=pTarget->nChannels) {
    if (1!=pSource->nChannels||2!=pTarget->nChannels) {
      // We should never go here ...
      DVMESSAGE("matrix conversion %u channels -> %u channels not supported",
          pSource->nChannels,pTarget->nChannels);
      goto error;
    }

    if (pSource->wBitsPerSample<pTarget->wBitsPerSample)
      *pFormat=*pTarget;
    else {
      *pFormat=*pSource;
      pFormat->nChannels<<=1u;
      pFormat->nBlockAlign<<=1u;
      pFormat->nAvgBytesPerSec<<=1ul;
    }

    pRing->Write=RingWriteSamplesDuplicate;
  }
  else {
    if (pSource->wBitsPerSample<pTarget->wBitsPerSample)
      *pFormat=*pTarget;
    else
      *pFormat=*pSource;

    pRing->Write=RingWriteSamples;
  }

  pRing->Read=RingReadSamples;

  /////////////////////////////////////////////////////////////////////////////
  switch (pEngine->options.common.eDither) {
  case eDitherRound:
    pRing->dither.lpVtbl=&gcIRingDitherRoundVtbl;
    break;
  case eDitherSimple:
    pRing->dither.lpVtbl=&gcIRingDitherSimpleVtbl;
    break;
  case eDitherTruncate:
  default:
    pRing->dither.lpVtbl=&gcIRingDitherNullVtbl;
    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  dwRingSize=pFormat->nBlockAlign*uFrames;

  /////////////////////////////////////////////////////////////////////////////
  if (!(pRing->_buf=malloc(dwRingSize))) {
    DMESSAGE("allocation ring");
    goto error;
  }

  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;
  pRing->_mp=pRing->_buf+dwRingSize;
  pRing->frames._uWritten=0u;
  pRing->frames._uAvailable=uFrames;

  DVWRITELN("uAvailable=%u uWritten=%u",pRing->frames._uAvailable,
      pRing->frames._uWritten);
  code=0;
error:
  return code;
}

void RingDestroy(Ring *pRing)
{
  if (pRing->_buf) {
    free(pRing->_buf);
    pRing->_buf=NULL;
  }

  pRing->pEngine=NULL;
}

int RingReallocFrames(Ring *pRing, Engine *pEngine, UINT32 uFrames)
{
  int code=-1;
  WAVEFORMATEX *pFormat=&pRing->wfx.Format;
  UINT32 uFramesRing=(pRing->_mp-pRing->_buf)/pFormat->nBlockAlign;
  Ring ring;

DVWRITELN("uFrames: %u, uFramesRing: %u ===============================",
    uFrames,uFramesRing);

  if (uFramesRing==uFrames)
    RingResetWrite(pRing,0);
  else {

    ring=*pRing;

    if (RingCreateFrames(pRing,pEngine,uFrames)<0) {
      DMESSAGE("creating ring");
      goto exit;
    }

    if (0u<pRing->frames._uWritten) {
      if (ring._wp<ring._rp) {
        memcpy(pRing->_wp,ring._rp,ring._mp-ring._rp);
        pRing->_wp+=ring._mp-ring._rp;
        DASSERT(pRing->_wp<pRing->_mp,exit);

        memcpy(pRing->_wp,ring._buf,ring._wp-ring._buf);
        pRing->_wp+=ring._wp-ring._buf;
        DASSERT(pRing->_wp<pRing->_mp,exit);
      }
      else {
        memcpy(pRing->_wp,ring._rp,ring._wp-ring._rp);
        pRing->_wp+=ring._wp-ring._rp;
        DASSERT(pRing->_wp<pRing->_mp,exit);
      }

      pRing->frames._uWritten=(pRing->_wp-pRing->_buf)/pFormat->nBlockAlign;
      pRing->frames._uAvailable=(pRing->_mp-pRing->_wp)/pFormat->nBlockAlign;
    }

    RingDestroy(&ring);
  }

  code=0;
exit:
  return code;
}

void RingReset(Ring *pRing)
{
  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {
  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;
  pRing->frames._uAvailable+=pRing->frames._uWritten;
  pRing->frames._uWritten=0u;
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return;
}

////////////////////////////////
int RingResetWrite(Ring *pRing, int bInitial)
{
  int code=-1;
  Engine *pEngine=pRing->pEngine;
  EMono2Stereo eMono2Stereo=pEngine->options.common.eMono2Stereo;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;
  WAVEFORMATEX *pTarget=&pEngine->wfx.target.Format;
  WAVEFORMATEXTENSIBLE *pWfx=&pRing->wfx;

  if (eMono2Stereo&&pSource->nChannels!=pTarget->nChannels) {
    if (1!=pSource->nChannels||2!=pTarget->nChannels) {
      // We should never go here ...
      DVMESSAGE("matrix conversion %u channels -> %u channels not supported",
          pSource->nChannels,pTarget->nChannels);
      goto error;
    }

    WfxCreate(pWfx,pSource->nSamplesPerSec,2,eRingFormatBitsPerSample,
        pEngine->options.common.bSurround);

    pRing->Write=RingWriteSamplesDuplicate;
  }
  else {
    WfxCreate(pWfx,pSource->nSamplesPerSec,pSource->nChannels,
        eRingFormatBitsPerSample,pEngine->options.common.bSurround);

    pRing->Write=RingWriteSamples;
  }

  code=0;
error:
  return code;
}

WORD RingFramesMaximum(Ring *pRing)
{
  return (pRing->_mp-pRing->_buf)/pRing->wfx.Format.nBlockAlign;
}

int RingWrite(Ring *pRing, const BYTE *buf, DWORD dwSizeSource)
{
  int code=-1;
  const WAVEFORMATEX *pSource=&pRing->pEngine->wfx.source.Format;
  const WAVEFORMATEX *pFormat=&pRing->wfx.Format;
  WORD uFrames=dwSizeSource/pSource->nBlockAlign;
  DWORD dwSize=pFormat->nBlockAlign*uFrames;
  DWORD dwSize1,dwSize2;

  /////////////////////////////////////////////////////////////////////////////
  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  /////////////////////////////////////////////////////////////////////////////
  EnterCriticalSection(&pRing->pEngine->cs); // {

  if (pRing->_wp<pRing->_rp) {
    // already wrapped
    if (pRing->_rp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto error;
    }

    DASSERT(0u==dwSize%pFormat->nBlockAlign,error);
    pRing->Write(pRing,&buf,dwSize);
  }
  else if (pRing->_mp<pRing->_wp+dwSize) {
    // will wrap
    dwSize1=pRing->_mp-pRing->_wp;
    dwSize2=dwSize-dwSize1;

    if (pRing->_rp<pRing->_buf+dwSize2) {
      DMESSAGE("overflow");
      goto error;
    }

    DASSERT(0u==dwSize1%pFormat->nBlockAlign,error);
    pRing->Write(pRing,&buf,dwSize1);
    DASSERT(0u==dwSize2%pFormat->nBlockAlign,error);
    pRing->Write(pRing,&buf,dwSize2);
  }
  else {
    // will not wrap
    if (pRing->_mp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto error;
    }

    DASSERT(0u==dwSize%pFormat->nBlockAlign,error);
    pRing->Write(pRing,&buf,dwSize);
  }

  pRing->frames._uAvailable-=uFrames;
  pRing->frames._uWritten+=uFrames;
  code=0;
error:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}

int RingReadFrames(Ring *pRing, BYTE *buf, WORD uFrames)
{
  int code=-1;
  const WAVEFORMATEX *pFormat=&pRing->wfx.Format;
  DWORD dwSize=pFormat->nBlockAlign*uFrames;
  DWORD dwSize1,dwSize2;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {

  if (pRing->_rp<pRing->_wp) {
    // not wrapped
    if (pRing->_wp<pRing->_rp+dwSize) {
      DMESSAGE("underflow");
      goto eunderflow;
    }

    pRing->Read(pRing,&buf,dwSize);
  }
  else {
    // wrapped
    dwSize1=pRing->_mp-pRing->_rp;

    if (dwSize<dwSize1)
      pRing->Read(pRing,&buf,dwSize);
    else {
      dwSize2=dwSize-dwSize1;
      pRing->Read(pRing,&buf,dwSize1);
      pRing->Read(pRing,&buf,dwSize2);
    }
  }

  pRing->frames._uWritten-=uFrames;
  pRing->frames._uAvailable+=uFrames;
  code=0;
eunderflow:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}

////////////////////////////////
DWORD RingSourceSize(Ring *pRing)
{
  WAVEFORMATEX *pSource=&pRing->pEngine->wfx.source.Format;
  WAVEFORMATEX *pFormat=&pRing->wfx.Format;

  return MulDiv(pRing->_mp-pRing->_buf,pSource->nBlockAlign,
      pFormat->nBlockAlign);
}

DWORD RingSourceSizeAvailable(Ring *pRing)
{
  WAVEFORMATEX *pSource=&pRing->pEngine->wfx.source.Format;

  return pSource->nBlockAlign*pRing->frames._uAvailable;
}

DWORD RingSourceSizeWritten(Ring *pRing)
{
  WAVEFORMATEX *pSource=&pRing->pEngine->wfx.source.Format;

  return pSource->nBlockAlign*pRing->frames._uWritten;
}

////////////////////////////////
DWORD RingTargetSize(Ring *pRing)
{
  WORD uFrames=(pRing->_mp-pRing->_buf)/pRing->wfx.Format.nBlockAlign;

  return pRing->pEngine->wfx.target.Format.nBlockAlign*uFrames;
}

DWORD RingTargetSizeWritten(Ring *pRing)
{
  WAVEFORMATEX *pTarget=&pRing->pEngine->wfx.target.Format;

  return pTarget->nBlockAlign*pRing->frames._uWritten;
}
