/*
 * yasapi_wfx.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

int WfxCreate(WAVEFORMATEXTENSIBLE *pWfx, DWORD samplerate, WORD numchannels,
    WORD bitspersamp, int bSurround)
{
  enum {
    eSize=(sizeof *pWfx)-sizeof (pWfx->Format),
  };

  int code=-1;

  pWfx->Format.wFormatTag=WAVE_FORMAT_EXTENSIBLE;
  pWfx->Format.cbSize=eSize;
  pWfx->SubFormat=KSDATAFORMAT_SUBTYPE_PCM;

  switch (numchannels) {
  case 1:
    pWfx->dwChannelMask=KSAUDIO_SPEAKER_MONO;
    DWRITELN("KSAUDIO_SPEAKER_MONO");
    break;
  case 2:
    pWfx->dwChannelMask=KSAUDIO_SPEAKER_STEREO;
    DWRITELN("KSAUDIO_SPEAKER_STEREO");
    break;
  case 3:
    pWfx->dwChannelMask
        =SPEAKER_FRONT_LEFT
        |SPEAKER_FRONT_RIGHT
        |SPEAKER_LOW_FREQUENCY;
    DWRITELN("SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT|SPEAKER_LOW_FREQUENCY");
    break;
  case 4:
    if (bSurround) {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_SURROUND;
      DWRITELN("KSAUDIO_SPEAKER_SURROUND");
    }
    else {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_QUAD;
      DWRITELN("KSAUDIO_SPEAKER_QUAD");
    }

    break;
  case 5:
    pWfx->dwChannelMask
        =SPEAKER_BACK_RIGHT
        |SPEAKER_FRONT_RIGHT
        |SPEAKER_FRONT_CENTER
        |SPEAKER_FRONT_LEFT
        |SPEAKER_BACK_LEFT;
    DWRITELN("SPEAKER_BACK_RIGHT|SPEAKER_FRONT_RIGHT|SPEAKER_FRONT_CENTER"
        "|SPEAKER_FRONT_LEFT|SPEAKER_BACK_LEFT");
    break;
  case 6:
    if (bSurround) {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_5POINT1_SURROUND;
      DWRITELN("KSAUDIO_SPEAKER_5POINT1_SURROUND");
    }
    else {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_5POINT1;
      DWRITELN("KSAUDIO_SPEAKER_5POINT1");
    }

    break;
  case 8:
    if (bSurround) {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_7POINT1_SURROUND;
      DWRITELN("KSAUDIO_SPEAKER_7POINT1_SURROUND");
    }
    else {
      pWfx->dwChannelMask=KSAUDIO_SPEAKER_7POINT1;
      DWRITELN("KSAUDIO_SPEAKER_7POINT1");
    }

    break;
  default:
    DVMESSAGE("%d channels not supported",numchannels);
    goto exit;
  }

  ////
  pWfx->Format.nSamplesPerSec=samplerate;
  pWfx->Format.nChannels=numchannels;
  pWfx->Format.wBitsPerSample=bitspersamp;
  DVWRITELN("*** samplerate=%lu numchannels=%u bitspersamp=%u",
      samplerate,numchannels,bitspersamp);
  ////
  pWfx->Format.nBlockAlign=pWfx->Format.nChannels;
  pWfx->Format.nBlockAlign*=pWfx->Format.wBitsPerSample>>3;

  ////
  pWfx->Format.nAvgBytesPerSec=pWfx->Format.nSamplesPerSec;
  pWfx->Format.nAvgBytesPerSec*=pWfx->Format.nBlockAlign;
  ////
  pWfx->Samples.wValidBitsPerSample=pWfx->Format.wBitsPerSample;

  ////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}
